from ._TableError import TableError


class TableIsNotInTabsMode(TableError):
    def __init__(self, table_id):
        self.table_id = table_id

    def get_extra_data(self):
        return {
            "tableId": self.table_id
        }
