from ._TableError import TableError


class TabIsNotASplitSeat(TableError):

    def __init__(
        self,       # type: TabIsNotASplitSeat
        tab_name,   # type: str
    ):
        # type: (...) -> None

        self.tab_name = tab_name

    def get_extra_data(self):
        return {
            "tabName": self.tab_name,
        }
