class Event(object):
    def __init__(
        self,
        event_id,           # type: str
        alert_id,           # type: str
        notification_json,  # type: str
        timestamp,          # type: str
        sent_status,        # type: str
        status_desc,        # type: str
    ):
        # type: (...) -> None

        self.id = event_id
        self.alert_id = alert_id
        self.notification_json = notification_json
        self.timestamp = timestamp
        self.sent_status = sent_status
        self.status_desc = status_desc
