class Table(object):

    def __init__(
        self,                       # type: Table
        id="",                      # type: str
        name="",                    # type: str
        current_pos_id="",          # type: str
        status="",                  # type: str
        business_period="",         # type: str
        sector="",                  # type: str
        slice_parent_table_id="",   # type: str
        slice_source_table_id="",   # type: str
        mode="",                    # type: str
        sub_sector="",              # type: str
        user_id="",                 # type: str
        sub_sector_name="",         # type: str
    ):
        # type: (...) -> None

        self.id = id
        self.name = name
        self.current_pos_id = current_pos_id
        self.status = int(status) if status != "" else None
        self.business_period = business_period
        self.sector = sector
        self.slice_source_table_id = slice_source_table_id
        self.slice_parent_table_id = slice_parent_table_id
        self.mode = mode
        self.sub_sector = sub_sector
        self.user_id = user_id
        self.sub_sector_name = sub_sector_name
