from ._TableError import TableError


class TabToJoinIsTheSame(TableError):

    def __init__(
        self,       # type: TabToJoinIsTheSame
    ):
        # type: (...) -> None
        super(TabToJoinIsTheSame, self).__init__("$SELECT_A_DIFFERENT_TAB")

    def get_extra_data(
        self,   # type: TabToJoinIsTheSame
    ):
        # type: (...) -> None

        return None
