from production.model import PicklistCleanerManagerConfig
from typing import List


class ProductionConfig(object):
    def __init__(
        self,                                   # type: ProductionConfig
        boxes_config,                           # type: List[object]
        branches_config,                        # type: List[object]
        views_config,                           # type: object
        picklist_database_manager_config,       # type: PicklistCleanerManagerConfig
    ):
        # type: (...) -> None

        self.boxes_config = boxes_config
        self.branches_config = branches_config
        self.views_config = views_config
        self.picklist_database_manager_config = picklist_database_manager_config
