class Statistics(object):
    def __init__(self, name, variable, format, goal, comparison):
        # type: (unicode, unicode, unicode, unicode, unicode) -> None

        self.name = name
        self.variable = variable
        self.format = format
        self.goal = goal
        self.comparison = comparison

        self.value = None
