# -*- coding: utf-8 -*-

class StoreNotOperatorWithSatException(Exception):
    def __init__(
        self,       # type: StoreNotOperatorWithSatException
    ):
        # type: (...) -> None

        self.message = "Loja não opera com SAT"
