
from typing import List


class OrderLines(object):
    def __init__(
        self,           # type: OrderLines
        order_id,       # type: int
        line_numbers,   # type: List[int]
    ):
        # type: (...) -> None

        self.order_id = order_id
        self.line_numbers = line_numbers
