import json


class PosUpdaterDto(object):

    def __init__(
        self,       # type: PosUpdaterDto
        pos_id,     # type: int
        user_id,    # type: str
    ):
        # type: (...) -> None

        self.pos_id = pos_id
        self.user_id = user_id

    def __str__(
        self,   # type: PosUpdaterDto
    ):
        # type: (...) -> str

        return json.dumps({
            "pos_id": self.pos_id,
            "user_id": self.user_id,
        })
