"""

Revision ID: 37f226415b7f
Revises: 2f33894eaa09
Create Date: 2024-11-27 10:28:21.831752

"""
from typing import (
    Sequence,
    Union,
)

from alembic import op
from sqlalchemy import (
    Column,
    DATETIME,
    TEXT,
    VARCHAR,
)

# revision identifiers, used by Alembic.
revision: str = "37f226415b7f"
down_revision: Union[str, None] = "2f33894eaa09"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    op.drop_column("Users", "PasswordValidity")
    op.drop_column("Users", "PayRate")
    op.drop_column("Users", "BadgeNumber")


def downgrade() -> None:

    op.add_column("Users", Column("PasswordValidity", DATETIME))
    op.add_column("Users", Column("PayRate", VARCHAR(30)))
    op.add_column("Users", Column("BadgeNumber", TEXT))
