from production.command import ProductionCommand


class PublishOrderCommand(ProductionCommand):
    def __init__(self, order_id):
        super(PublishOrderCommand, self).__init__(order_id, None)

    def get_hash_extra_value(self):
        return "PublishOrderCommand"
