class FiscalInfo(object):

    def __init__(
        self,               # type: FiscalInfo
        nf_type,            # type: str
        initial_order_id,   # type: int
    ):
        # type: (...) -> None

        self.nf_type = nf_type
        self.initial_order_id = initial_order_id
