class SaleLineIdentifier(object):

    def __init__(
        self,               # type: SaleLineIdentifier
        line_number=None,   # type: int
        item_id=None,       # type: str
        level=None,         # type: int
        part_code=None,     # type: int
    ):
        # type: (...) -> None

        self.line_number = line_number
        self.item_id = item_id
        self.level = level
        self.part_code = part_code
