from typing import Optional


class AsyncTarget(object):

    def __init__(
        self,               # type: AsyncTarget
        reply_subject,      # type: str
        id_field=None,      # type: Optional[str]
    ):
        self.reply_subject = reply_subject
        self.id_field = id_field or "id"
