class CannotTotalEmptyOrder(Exception):

    def __init__(
        self,       # type: CannotTotalEmptyOrder
    ):
        # type: (...) -> None
        super(CannotTotalEmptyOrder, self).__init__()
