from typing import Dict


class TranslatedMessage(object):
    def __init__(
        self,       # type: TranslatedMessage
        pt,         # type: str
        en,         # type: str
        es,         # type: str

    ):
        # type: (...) -> None

        self.en = en
        self.es = es
        self.pt = pt
